# =========================================================================
#
# eq_event
#
# NOTE: this just does a comparison of two 'labels':
# it compares text strings, string that look like lotos events.
# it does a case insensitive string comparison, after removal of
# whitespace.
# examples:
#    eq_event "obuin ! M1 ! PUR4 ! C1"  "obuin ! M1 ! PUR0 ! C3"
# 0
#    eq_event "obuin ! M1 ! PUR4 ! C1"  "obuin ! M1 ! PUR4 ! C3"
# 0
#    eq_event "obuin ! M1 ! PUR4 ! C1"  "obuin ! M1 ! PUR4 ! c1"
# 1
#

proc eq_event {event_a event_b} {
	# try to remove whitespace and empty parentheses
	regsub -all -- "\t| |\\(\\)" $event_a "" event_a
	regsub -all -- "\t| |\\(\\)" $event_b "" event_b
	set event_a [string tolower $event_a]
	set event_b [string tolower $event_b]
	if {[string compare $event_a $event_b] == 0} {
		return 1
	} else {
		return 0
	}
}

# =========================================================================

# transform a string representaion of an event to a hierarchical list
# that represents it as a list of terms.
# each node is represented by a (sub)list of two items:
#  - the 'operator' 
#  - the list of subnodes (can be the empty list {})
# Each element in the list of subnodes is a node, i.e. a list of two
# elements: the 'operator' and the 'subterms'.
#
# Examples:
# "g ! a"                => {g {}} {a {}} 
# "g ! a ! b"            => {g {}} {a {}}  {b {}} 
# "g ! a(b,d)"           => {g {}} {a {{b {}} {d {}}}} 
# "g ! a(b(c,d,e),f)"    => {g {}} {a {{b {{c {}} {d {}} {e {}}}} {f {}}}} 
# "g ! a(b(c,d(h),e),f)" => {g {}} {a {{b {{c {}} {d {{h {}}}} {e {}}}} {f {}}}}
#
# NOTE: also the gate will have an (always empty) list of subterms!

proc event2term { e } {
	# first, do some checking on the arguments:
	# - we have to have as much { as }
	set co [regsub -all {\{} $e d bogus ]
	set cc [regsub -all {\}} $e d bogus ]
	if {$co != $cc} {
		error {event2term: unbalanced { and } in input}
	}
	# get rid of whitespace:
	# (we need "" around the match-expression here,
	# because we want \t to be expanded to a TAB character, which will
	# not happen if we would put {} around the expression.
	regsub -all "\[ \t]" $e "" e2
	# add 'missing' parenthesis around whole expression,
	# and replace '!' by ')(', to make each gate or experiment offer
	# into a sublist (next two commands):
	set e3 "($e2)"
	regsub -all {!} $e3 {)(} e4
	# protect all embedded $'s in the event
	regsub -all \\\$ $e4 \\\$ e4
	# protect expressions within { and } (representations of sets and the like)
	# we do this by replacing them by variables, and setting the variables
	# to those expressions. The variables will be expanded again by the final
	# subst, and then the list command around it will take care of protecting
	# the { }
	set i 1
	set eee $e4
	while {[regexp "(\{\[^\}]*\})" $eee all match]} {
		regsub "(\{\[^\}]*\})" $eee "\$p$i" eee
		set p$i $match
		incr i
	}
	# now, do some checking on the arguments:
	# - we have to have as much ( as )
	set co [regsub -all {\(} $eee d bogus ]
	set cc [regsub -all {\)} $eee d bogus ]
	if {$co != $cc} {
		error {event2term: unbalanced ( and ) in input}
	}
	# first get rid of empty parentheses
	# (we will add them ourselves where necessary!)
	regsub -all {\(\)} $eee "" e5
	# double the existing parentheses
	regsub -all {\(} $e5 {((} e5a
	regsub -all {\)} $e5a {))} e5b
	set e5 $e5b
	# add 'missing' parenthesis after nullary operators:
	regsub -all {([^()])([!,)])} $e5 {\1()\2} e6
	### puts $e6
	# get rid of the (now) no longer necessary ',':
	regsub -all {,} $e6 ")(" e7
	### puts $e7
	# replace '(' and ')' by ' [list ' resp ' ] ' (in next two commands):
	regsub -all {\(} $e7 { [list } e8
	regsub -all {\)} $e8 { ] } e9
	### puts $e9
	set t [subst $e9]
	return $t
}

# =========================================================================

# print a term given in above tree/list representation.
proc printterm_plain {term {indentation ""}} {
	foreach node $term {
		set operator [lindex $node 0]
		set subtreelist [lindex $node 1]
		puts "$indentation$operator"
		printterm_plain $subtreelist "$indentation "
	}
}

# =========================================================================
#
# treewalk routine that takes:
#  - a list (e.g. the result of event2term)
#  - a 'loop variable name' to which the operator names will be bound
#  - a 'loop variable name' that will set to the 'mode' (enter or leave)
#  - a 'loop variable name' that will set to 'path' of the current node
#  - a 'body' of tcl code that is to be executed for each tree node,
#    when entering the node (mode set to enter) and, again, when leaving
#    the tree node (mode set to leave).
#
# example of usage:
#
# visit [event2term "g!a(b,c(d))"] op m p {
#	puts "$op $m $p"
# }
#
# result:
#
# g enter 0
# g leave 0
# a enter 1
# b enter 1.0
# b leave 1.0
# c enter 1.1
# d enter 1.1.0
# d leave 1.1.0
# c leave 1.1
# a leave 1
#
proc visit {treelist nameVar modeVar pathVar args} {
	# take body from args, and pass it as a separate argument to do_visit,
	# because otherwise the vars and the body will end up in a single list
	set body [lindex $args end]
	set vars [lrange $args 0 [expr [llength $args] - 2]]
	
	# make the vars known in this scope
	foreach v $vars {
		# puts stderr "upvar-ing $v $v"
		upvar $v $v
	}
	
	# set son-nr to 0 -- it will be incremented for each term/son we look at
	set son 0
	# call the routine that does the real work for each 'subtree' in the list
	# make sure to pass $vars and $body as separate arguments: NOT as $args!!!
	foreach tree $treelist {
		set suboperator [lindex $tree 0]
		set subtreelist [lindex $tree 1]
		do_visit $suboperator $subtreelist $nameVar $modeVar $pathVar $son $vars $body
		incr son
	}
}

# do_visit is a routine used by visit.
# It is NOT to be invoked directly by the user
proc do_visit {operator treelist nameVar modeVar pathVar path vars body} {
	# make sure that the variables are known when we 'uplevel $body'
	upvar $nameVar name
	upvar $modeVar mode
	upvar $pathVar pathv
	
	# take body from args, and pass it as a separate argument to do_visit,
	# because otherwise the vars and the body will end up in a single list
	# set body [lindex $args end]
	# set vars [lrange $args 0 [expr [llength $args] - 2]]

	# make the vars known in this scope
	foreach v $vars {
		# puts stderr "upvar-ing $v $v"
		upvar $v $v
	}

	# set the 'loop' variables to the right values
	set pathv $path
	set name $operator

	# set the mode 'loop' variable to 'enter', and execute the body
	set mode enter
	uplevel $body

	# set son-nr to 0 -- it will be incremented for each son we look at
	set son 0
	# recursively call the routine that does the real work;
	# make sure to pass $vars and $body as separate arguments: NOT as $args!!!
	foreach tree $treelist {
		set suboperator [lindex $tree 0]
		set subtreelist [lindex $tree 1]
		do_visit $suboperator $subtreelist $nameVar $modeVar $pathVar $path.$son $vars $body 
		incr son
	}

	# set the mode 'loop' variable to 'leave', and execute the body
	set mode leave
	uplevel $body

}
# =========================================================================

# print a term using above visit routine
#
proc printterm {term {indentation ""}} {
	visit $term operator m p indentation {
		if {[string match $m  enter]} {
			puts "$indentation$operator"
			append indentation " "
		}
		if {[string match $m leave]} {
			regsub { $} $indentation "" indentation
		}
	}
}

# =========================================================================

# convert a term back to an event using above visit routine
# the result is a list of experiments (gate is seen as first experiment)
#
# Cases to test:
# join [term2event [event2term "g ! a(b(j,k,l),c(d,e),f) ! h" ] ] " ! "
# join [term2event [event2term "g ! a(b(j(k(l(m)))),c(d,e),f) ! h" ] ] " ! "
#
proc Debug {s} {
	global maininfo
	if {[info exists maininfo(debug)]} {
		if {$maininfo(debug)} {
			puts stderr $s
		}
	}
}
proc term2event {term} {
	set level 0
	set el_stack {}
	set nr_stack {}
	set indentation ""
	
	visit $term operator m p indentation level el_stack nr_stack {
	    Debug "I<$operator><$m><$p><$indentation><$level><$el_stack><$nr_stack>"
		if {[string match $m  enter]} {
			Debug "$indentation$operator"
			append indentation " "
			incr level
		}
		if {[string match $m leave]} {
			regsub { $} $indentation "" indentation
		  if {[llength $nr_stack] > 0} {
			set nr [lindex $nr_stack end]
			if {$nr > $level} {
				set subs [lindex $el_stack end]
				if {[llength $subs] > 0} {
					# be sure NOT to introduce whitespace here
					# because inotrucing whitespace
					# will cause problems in the list handling!!!
					set subs_str [join $subs ","]
					set el_to_add "[set operator]($subs_str)"
				} else {
					set el_to_add $operator
				}
				set el_stack [lreplace $el_stack end end]
				set nr_stack [lreplace $nr_stack end end]
				if {[llength $nr_stack] > 0} {
					set nr_on_stack [lindex $nr_stack end]
					if {$nr_on_stack == $level} {
						if {[llength $el_stack] > 0} {
							set brothers [lindex $el_stack end]
							set el_stack [lreplace $el_stack end end]
						} else {
							set brothers {}
						}
						lappend brothers $el_to_add
						lappend el_stack $brothers
					} elseif {$nr_on_stack < $level} {
						lappend nr_stack $level
						lappend el_stack $el_to_add
					}
				} else {
					lappend nr_stack $level
					lappend el_stack $el_to_add
				}
			} elseif {$nr == $level} {
				if {[llength $el_stack] > 0} {
					set brothers [lindex $el_stack end]
					set el_stack [lreplace $el_stack end end]
				} else {
					set brothers {}
				}
				lappend brothers $operator
				lappend el_stack $brothers
			} elseif {$nr < $level} {
				lappend el_stack $operator
				lappend nr_stack $level
			}
		  } else {
			lappend nr_stack $level
			lappend el_stack $operator
		  }
		  incr level -1
		}
	    Debug "O<$operator><$m><$p><$indentation><$level><$el_stack><$nr_stack>"
	}
	return [lindex $el_stack end]
}

# =========================================================================

proc path2operator {term path} {
	set poslist [split $path .]
	set sub $term
	set op ""
	foreach p $poslist {
		set op [lindex [lindex $sub $p] 0]
		set sub [lindex [lindex $sub $p] 1]
	}
	Debug "path2operatorpos ($term)($path): ($op)"
	return $op
}

# =========================================================================

proc path2arity {term path} {
	set poslist [split $path .]
	set sub $term
	foreach p $poslist {
		set sub [lindex [lindex $sub $p] 1]
	}
	set res [llength $sub]
	Debug "path2arity ($term)($path): ($res)"
	return $res
}

# =========================================================================

proc path2tuple {term path} {
	set poslist [split $path .]
	set sub $term
	set op ""
	foreach p $poslist {
		set op [lindex [lindex $sub $p] 0]
		set sub [lindex [lindex $sub $p] 1]
	}
	set res [list $op [llength $sub]]
	Debug "path2tuple ($term)($path): ($res)"
	return $res
}

# =========================================================================
