# =========================================================================
# $Id: io.tcl,v 1.1.2.21 2004/08/30 17:50:18 belinfan Exp $
# =========================================================================
# IO wrapper routines
# Interface routines:
#
# io_send_command peer type <fieldname,value>-list
# io_send_command peer -multiline-start type
# io_send_command peer -multiline-end type
# io_send_command_line peer type line
# io_send_answer peer type <fieldname,value>-list
# io_send_answer peer -multiline-start type
# io_send_answer peer -multiline-end type
# io_send_answer_line peer type line
#
# io_receive_command peer typed-answer-clauses-list
# io_receive_answer peer typed-answer-clauses-list
# where
# typed-answer-clause == -multiline type answer-clauses-list
# typed-answer-clause == -singleline answer-clauses-list
# answer-clause == type fieldname-list command
#   where command is executed in a scope in which for each
#   fieldname of fieldname-list there is a variable fieldname
#   which is set to the value present in the answer
#   If no fields are specified, the variable 'line' contains
#   the whole line that was matched (except trailing \r\n).
# answer-clause == 'timeout' command
#   executed when timeout occurs.
# answer-clause == {} command
#   executed when none of the other clauses match.
#   It matches a single line of text.
#   The variable 'line' contains the whole line
#   that was matched (except trailing \r\n).
# answer-clause == 'default' command
#   executed when none of the other clauses match, and no
#   default (expect_before/after) rule is set.
# answer-clause == 'start'  command
#   executed when the start of a multi-line answer has been seen
#   this will be the first one executed for a multi-line answer
# answer-clause == 'end'  command
#   executed when the last element of a multi-line answer has been seen
#   this will be the last one executed for a multi-line answer
#
# Note: there is an automatically generated receive clause that removes
# echo-ed lines that we sent. It removes those lines by matching them
# with resp. "^C_[^\r\n]*\r*\n" and "^A_[^\r\n]*\r*\n" .
#
# io_config timeout ....?
#
# io_add_peer peer spawnid timeout ....?
#
# io_rm_peer peer
#

proc io_init {} {
	global io
	global timeout
	if {[info exists timeout]} {
		set io(timeout) $timeout
	} else {
		set io(timeout) 10
	}
	# set io(nl) "\r"
	set io(nl) "\n"
}

proc io_config {timeout} {
	global io
	set io(timeout) $timeout
}

proc io_add_peer {peer spawnid timeout} {
	global io
	io_debug 2 "io_add_peer $peer $spawnid $timeout"
	set io(sid,$peer) $spawnid
	set io(timeout,$peer) $timeout
}

proc io_rm_peer {peer} {
	global io
	catch {unset io(sid,$peer)}
	catch {unset io(timeout,$peer)}
}

proc io_send_command {peer type {name_value_list {}} } {
	global io
	if {[info exists io(sid,$peer)]} {
		io_do_send $peer C_ $type $name_value_list
	}
}
proc io_send_answer {peer type {name_value_list {}} } {
	global io
	if {[info exists io(sid,$peer)]} {
		io_do_send $peer A_ $type $name_value_list
	}
}

proc io_send_command_line {peer type line} {
	global io
	if {[info exists io(sid,$peer)]} {
		io_do_send_line $peer C_ $type $line
	}
}
proc io_send_answer_line {peer type line} {
	global io
	if {[info exists io(sid,$peer)]} {
		io_do_send_line $peer A_ $type $line
	}
}

proc io_receive_command {peer clause_list} {
	global io
	if {[info exists io(sid,$peer)]} {
		set cmd [io_compile_receive $peer C_ A_ $clause_list]
		io_debug 2  "CMD: ($cmd)"
		uplevel 1 $cmd
	}
}

proc io_receive_answer {peer clause_list} {
	global io
	if {[info exists io(sid,$peer)]} {
		set cmd [io_compile_receive $peer A_ C_ $clause_list]
		io_debug 2  "CMD: ($cmd)"
		uplevel 1 $cmd
	}
}

# =========================================================================

proc io_do_send {peer prefix type {name_value_list {}} } {
	global io
	if {[info exists io(sid,$peer)]} {
		if {[string compare $type -multiline-start] == 0} {
			set realtype [lindex $name_value_list 0]
			set suffix _START
			set str "[set prefix][set realtype][set suffix]"
		} elseif {[string compare $type -multiline-end] == 0} {
			set realtype [lindex $name_value_list 0]
			set suffix _END
			set str "[set prefix][set realtype][set suffix]"
		} else {
			set str "[set prefix][set type] [fields_join $name_value_list]"
		}
		send -i $io(sid,$peer) "$str[set io(nl)]"
		io_debug 1 "sent $peer \"$str\"nl"
	}
}
proc io_do_send_line {peer prefix type line} {
	global io
	if {[info exists io(sid,$peer)]} {
		set str "[set prefix][set type] $line"
		send -i $io(sid,$peer) "$str[set io(nl)]"
		io_debug 1 "sent $peer \"$str\"nl"
	}
}

proc io_compile_receive {p prefix ignore_prefix cl} {
	global io
	if {![info exists io(compiled,$p,$prefix,$ignore_prefix,$cl)]} {
		set code [io_do_compile_receive $p $prefix $ignore_prefix $cl]
		set proc ""
		set proc_name [io_get_gen_procname]
		append proc "proc $proc_name {} {\n"
		append proc "  uplevel 1 {\n"
		append proc $code
		append proc "\n"
		append proc " }\n"
		append proc "}\n"
		eval $proc
		set io(compiled,$p,$prefix,$ignore_prefix,$cl) $proc_name
	}
	return $io(compiled,$p,$prefix,$ignore_prefix,$cl)
}

proc io_get_gen_procname {} {
	global io
	if {![info exists io(gen,proc,nr)]} {
		set io(gen,proc,nr) 0
	}
	incr io(gen,proc,nr)
	return "io_gen_proc_[set io(gen,proc,nr)]"
}

proc io_do_compile_receive {p pfx ign_pfx cl} {
	global io
	set pcl [io_parse_clauses $cl]
	io_debug 3 $pcl
	set ml [lindex $pcl 0]
	set sl [lindex $pcl 1]
	set ml_t [lindex $ml 0]
	set ml_c [lindex $ml 1]
	set ml_def [lindex $ml 2]
	set ml_to [lindex $ml 3]
	set ml_start [lindex $ml 4]
	set ml_end [lindex $ml 5]
	set sl_c [lindex $sl 0]
	set sl_def [lindex $sl 1]
	set sl_to [lindex $sl 2]
	set ind "\t"
	set cmd ""
	append cmd "[set ind]global io\n"
	append cmd "[set ind]global timeout\n"
	append cmd "[set ind]set timeout \$io(timeout,$p)\n"
	append cmd "[set ind]io_debug_expect 1\n"
	append cmd "[set ind]expect {\n"
	set b_ind $ind
	append ind "\t"
	append cmd [io_compile_multiline $p $pfx $ign_pfx $ml_t $ml_start $ml_end $ml_c $ml_def $ml_to $ind]
	append cmd [io_compile_singleline $p $pfx $ign_pfx $sl_c $sl_def $sl_to $ind]
	append cmd [io_compile_multiline_def $p $pfx $ign_pfx $ml_t $ml_start $ml_end $ml_c $ml_def $ml_to $ind]
	set ind $b_ind
	append cmd "[set ind]}\n"
	append cmd "[set ind]io_debug_expect 0\n"
	append cmd "[set ind]set timeout \$io(timeout)\n"
	return $cmd
}
proc io_parse_clauses {clauses} {
	set ml {}
	set sl {}
	set i 0
	foreach clause $clauses {
		io_debug 3 "Clause($i):<<$clause>>"
		incr i
	}
	while {[llength $clauses] > 0} {
		set key [lindex $clauses 0]
		if {[string compare $key "-singleline"] == 0} {
			set slc [lindex $clauses 1]
			set sl [io_do_parse_clauses $slc]
			set clauses [lrange $clauses 2 end]
		} elseif {[string compare $key "-multiline"] == 0} {
			set mlc_t [lindex $clauses 1]
			set mlc_c [lindex $clauses 2]
			# NOTE: we lappend all elements of ml one by one
			# instead of simply using concat, because we want to
			# (no, we have to!) avoid expansion of variables in mlc_t. 
			set ml {}
			lappend ml $mlc_t
			foreach i [io_do_parse_clauses $mlc_c] {
				lappend ml $i
			}
			set clauses [lrange $clauses 3 end]
		} else {
			# error or sensible default
		}
	}
	return [list $ml $sl]
}
proc io_do_parse_clauses {clauses} {
	set c {}
	set def {}
	set to {}
	set end {}
	set start {}
	io_debug 3 "ParseClauses:<<$clauses>>"
	set lines [split $clauses \n]
	set cleanlines {}
	foreach l $lines {
		if {[regexp -- "^\[ \t\r\n\]*#" $l]} {
			io_debug 3  "ParseSkipClause:<<$l>>"
		} elseif {[regexp -- "^\[ \t\r\n\]*$" $l]} {
			io_debug 3  "ParseSkipClause:<<$l>>"
		} else {
			lappend cleanlines $l
		}
	}
	set clauses [join $cleanlines \n]
	io_debug 3 "ParseCleanClauses:<<$clauses>>"
	while {[llength $clauses] > 0} {
		set key [lindex $clauses 0]
		if {[string compare $key end] == 0} {
			set end [lindex $clauses 1]
			set clauses [lrange $clauses 2 end]
		} elseif {[string compare $key start] == 0} {
			set start [lindex $clauses 1]
			set clauses [lrange $clauses 2 end]
		} elseif {[string compare $key default] == 0} {
			set def [lindex $clauses 1]
			# next line would map 'default' onto '{}'
			# lappend c [list {} {} [lindex $clauses 2]]
			set clauses [lrange $clauses 2 end]
		} elseif {[string compare $key timeout] == 0} {
			set to [lindex $clauses 1]
			set clauses [lrange $clauses 2 end]
		} else {
			lappend c [list $key [lindex $clauses 1] [lindex $clauses 2]]
			set clauses [lrange $clauses 3 end]
		}
	}
	return [list $c $def $to $start $end]
}

proc io_compile_multiline {p pfx ign_pfx mlt start end cs def to ind} {
	global io
	set cmd ""
	if {[string compare $mlt ""] != 0} {
		append cmd "[set ind]-i \$io(sid,$p) -re \"^[set pfx]"
		append cmd $mlt
		append cmd "_START\\r*\\n\" {\n"
		set ind [io_indent $ind]
		append cmd $start
		append cmd "\n"
		append cmd "[set ind]expect {\n"
		set ind [io_indent $ind]
		append cmd [io_compile_singleline $p $pfx $ign_pfx $cs $def $to $ind $mlt $end]
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
	}
	return $cmd
}

proc io_compile_multiline_def {p pfx ign_pfx mlt start end cs def to ind} {
	global io
	set cmd ""
	if {[string compare $mlt ""] != 0} {
# will always be added by io_compile_singleline
#		if {[llength $ign_pfx] > 0} {
#			append cmd "[set ind]-i \$io(sid,$p) -re \"^[set ign_pfx]"
#			append cmd "\\\[^\\r]*\\r*\\n\""
#			append cmd " exp_continue\n"
#		}
# just use -singleline { timeout {bla} }
#		if {[llength $to] > 0} {
#			append cmd "[set ind]timeout {\n"
#			append cmd $to
#			append cmd "\n"
#			append cmd "[set ind]}\n"
#		}
		if {[llength $def] > 0} {
			append cmd "[set ind]default {\n"
			append cmd "$def\n"
			append cmd "\n"
			append cmd "[set ind]}\n"
		}
	}
	return $cmd
}

proc io_compile_singleline {p pfx ign_pfx cs def to ind {mlt ""} {end {}}} {
	global io
	set cmd {}
	if {[string compare $mlt ""] != 0 } {
		append cmd "[set ind]-i \$io(sid,$p)"
		append cmd " -re \"^[set pfx][set mlt]_END\\r*\\n\" {\n"
		append cmd $end
		append cmd "\n"
		append cmd "[set ind]}\n"
	}
	if {[llength $ign_pfx] > 0} {
		append cmd "[set ind]-i \$io(sid,$p) -re \"^[set ign_pfx]"
		append cmd "\\\[^\\r\\n]*\\r*\\n\""
		append cmd " exp_continue\n"
	}
	if {[llength $cs] > 0} {
		append cmd "[set ind]-i \$io(sid,$p)"
		append cmd " -notransfer -re \"^("
		set sep ""
		set pfxregexp ""
		set nonpfxregexp ""
		foreach c $cs {
			io_debug 3  "SingleLineClause:<<$c>>"
			if {[string compare [lindex $c 0] ""] != 0} {
				append pfxregexp $sep
				append pfxregexp "([set pfx][lindex $c 0])"
				set sep "|"
			} else {
				set nonpfxregexp "(\\\[\\t ]*\\\[^\\r\\n])"
			}
		}
		set sep ""
		if {[string compare $pfxregexp ""] != 0} {
			append cmd "(($pfxregexp)((\\\[\\t ]*\\\[^\\r\\n])|(\\r*\\n)))"
			set sep "|"
		}
		if {[string compare $nonpfxregexp ""] != 0} {
			append cmd $sep
			append cmd "($nonpfxregexp)"
		}
		append cmd ")\" {\n"
		set ind [io_indent $ind]
		append cmd "[set ind]set mybuf \"\"\n"
		append cmd "[set ind]expect {\n"
		set ind [io_indent $ind]
		append cmd "[set ind]-i \$io(sid,$p)"
		append cmd " -re"
		append cmd " \"^(\\\[^\\r\\n]\\\[^\\r\\n]*)\" {\n"
		set ind [io_indent $ind]
		append cmd "[set ind]append mybuf \$expect_out(1,string)\n"
		append cmd "[set ind]exp_continue\n"
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
		append cmd "[set ind]-i \$io(sid,$p)"
		append cmd " -re"
		append cmd " \"^(\\r*\\n)\" {\n"
		set ind [io_indent $ind]
		append cmd "[set ind]append mybuf \$expect_out(1,string)\n"
		set sep "[set ind]"
		set last ""
		foreach c $cs {
			append cmd "[set sep]if {\[regexp"
			if {[string compare [lindex $c 0] ""] != 0} {
				append cmd " \"^[set pfx][lindex $c 0](\\\[\\t ]\\\[\\t ]*(\\\[^\\r\\n]*))?\\r*\\n\""
			} else {
				append cmd " \"^(\\\[\\t ]*(\\\[^\\r\\n]*))?\\r*\\n\""
			}
			append cmd " \$mybuf match ws_rest rest\]}"
			append cmd " {\n"
			set ind [io_indent $ind]
			if {[llength [lindex $c 1]] > 0} {
				append cmd "[set ind]if {\[info exists rest]} {\n"
				set ind [io_indent $ind]
				append cmd "[set ind]set fields \[fields_split \$rest \]\n"
				set ind [io_outdent $ind]
				append cmd "[set ind]} else {\n"
				set ind [io_indent $ind]
				append cmd "[set ind]set fields {}\n"
				set ind [io_outdent $ind]
				append cmd "[set ind]}\n"
				foreach f [lindex $c 1] {
					append cmd "[set ind]set [set f] \[field_value \$fields [set f]\]\n"
				}
			} else {
				append cmd "[set ind]if {\[info exists rest]} {\n"
				set ind [io_indent $ind]
				append cmd "[set ind]set line \$rest\n"
				set ind [io_outdent $ind]
				append cmd "[set ind]} elseif {\[info exists ws_rest]} {\n"
				set ind [io_indent $ind]
				append cmd "[set ind]set line \$ws_rest\n"
				set ind [io_outdent $ind]
				append cmd "[set ind]} else {\n"
				set ind [io_indent $ind]
				append cmd "[set ind]set line \"\"\n"
				set ind [io_outdent $ind]
				append cmd "[set ind]}\n"
			}
			append cmd [lindex $c 2]
			append cmd "\n"
			set ind [io_outdent $ind]
			append cmd "[set ind]}"
			set sep " else"
			set last "\n"
		}
		append cmd $last
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
		if {[llength $to] > 0} {
			append cmd "[set ind]timeout {\n"
			append cmd "$to\n"
			append cmd "\n"
			append cmd "[set ind]}\n"
		}
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
		if {[string compare $mlt ""] != 0 } {
			append cmd "[set ind]exp_continue\n"
		}
		set ind [io_outdent $ind]
		append cmd "[set ind]}\n"
	}
	if {[llength $to] > 0} {
		append cmd "[set ind]timeout {\n"
		append cmd "$to\n"
		append cmd "\n"
		append cmd "[set ind]}\n"
	}
	if {[llength $def] > 0} {
		append cmd "[set ind]default {\n"
		append cmd "$def\n"
		append cmd "\n"
		append cmd "[set ind]}\n"
	}
	return $cmd
}

proc io_indent {ind} {
	return "[set ind]\t"
}

proc io_outdent {ind} {
	return [string range $ind 0 [expr [string length $ind] - 2]]
}

proc io_error {s} {
	global io
	puts stderr "io: $s"
}

proc io_debug {nr s} {
	global io
	if {[info exists io(debug)]} {
		if {$io(debug) >= $nr} {
			puts stderr "DEBUG $s"
		}
	}
}

proc io_debug_expect {val} {
	global io
	if {[info exists io(debug,expect)]} {
		if {$io(debug,expect)} {
			exp_internal $val
		}
	}
}

# =========================================================================
proc io_test {} {
	global aap
	global noot
#set aap [io_compile_receive spec OUTPUTS {
#		EVENT {event channel suspension} {
#			lappend events [list $channel $event $suspension]
#		}
#		end {set result [list $type $events ]}
#		default {set result [list error ]}
#	}
#]
	set aap [io_compile_receive spec A_ C_ {
		-multiline OUTPUTS {
			start {set starting [list $type $events ]}
			EVENT {event channel suspension} {
				lappend events [list $channel $event $suspension]
			}
			end {set result [list $type $events ]}
			{} {} {set result [list error_ml_empty ]}
			timeout {blo}
			default {set result [list error_ml_def ]}
		}
		-singleline {
			timeout {bla}
			default {set result [list error_sl_def ]}
		}
	}
	]
	set noot [io_compile_receive spec A_ C_ {
		-multiline STATS {
			STATS {} {
				lappend stats $line
			}
			end {set result [list stats $stats ]}
			timeout {set result [list error ]}
		}
		-singleline {
			STATS {} {set result [list stats [list $line] ]}
		}
	}
	]
	set mies [io_compile_receive spec A_ C_ {
		-multiline STATE {
			end {set result [list state $state ]}
			{} {} {
				lappend state $line
			}
		}
	}
	]
	set wim [io_compile_receive spec A_ C_ {
		-singleline {
			INPUT_OK {event channel suspension} {
				set result [list input_ok $channel $event $suspension]
			}
			INPUT_ERROR {reason} {
				set result [list input_error $reason]
			}
			INPUT {event channel suspension} {
				set result [list input_ok $channel $event $suspension]
			}
			"ERROR Inconsistency" {} {
				set result [list input_error $line]
			}
		}
	}
	]
	return "$aap $noot $mies $wim"
}

# =========================================================================
