# =========================================================================
#
# main loop for generating a fully exhaustive test suite in a braindead way
# 
# to improve:
#    the innermost loop over position should only do list replacement
#    using precomputed replacement sublist.
#    the replacement sublists can be computed in a loop more outer,
#    and can even be cached (using the StateSetRef as index)
#
# essential idea:
#
# given a suspension automaton (tree), in which we have
# an ordering over the outgoing edges (induced by ordering over the labels?)
# can we encode (linearise) a test case (a tree) as a sequence of choices?
# toplevel choices:
# Treenode ::=
# 	'PH' StateSetRef   % unexpanded
# |	'R' '0'	% reset
# |	'S' '1' StimId Treenode
# |	'O' SubtreeCount SubtreeCount*< ObsId Treenode >
# |	'P' '0'    % stop (pass)
# PH no choice yet (unexpanded, will be followed by state set identifier)
# S stimulate (will be followed by '1' (number of subtrees) and stimulus)
# O observe (will be followed by a number of number of valid edges and subtrees)
# P stop (pass, will be followed by '0' (number of subtrees))
# R reset (will be followed by '0' (number of subtrees))

# below we encode using numbers instead of PH S O P
#
# 0 S0    (no choice, state set)
# 1 0 S0 (stimulate, no choice, state set)
# 1 1 3 0 S1 (stimulate (1 out), stimulus 3, no choice, state set)
# 1 1 4 0 S2 (stimulate (1 out), stimulus 4, no choice, state set)
# 1 1 3 2 5 1 0 S3 2 0 S4 3 0  S5 7 0 S6 9 0 S7 (stimulate (1 out), stimulus 3, observe (5 valid out))
# 1 1 4 0 S2 (stimulate, stimulus 4, no choice, state set)

# five bags:
# - READY one of finished (closed) test cases
# - IN one of open test cases from which we pick
# - OUT one of open test cases that we produce in this step
# - CUR_IN, CUR_OUT what we are working on
# while not deep enough
#	while IN not empty
#		pick t from IN,
#		get placeholders into poslist
#		if poslist is empty
#			put t in READY
#		if poslist is notempty
#			put t in CUR_IN
#			for p in reverse(poslist) % to avoid expanding placeholders introduced at this level
#				while CUR_IN not empty
#					pick from CUR_IN
#					generate all possibilities for this p, and put all in CUR_OUT
#				move CUR_OUT to CUR_IN
#			move CUR_OUT and/or CUR_IN to OUT
#	move OUT to IN
#
# dependencies (for the encoding)
#  	L_I and L_U are known
#		(or become known on the fly by collecting new (previously unseen) items from spec)
# 	order over L_I and L_U  exists
# advantage:
#	only local knowledge is used in expansion
# disadvantage
#	less easy to use non-local knowledge in expansion (i.e. avoid observing after quiescence) 

proc get_ph_pos { t } {
	set l {}
	set i 0
	foreach e $t {
		if {[string compare $e PH] == 0} {
			lappend l $i
		}
		incr i
	}
	return $l
}

proc lreverse { t } {
	set l {}
	foreach e $t {
		set l [concat $e $l]
	}
	return $l
}

proc initcount { l } {
	set r {}
	foreach e $l {
		lappend r 0
	}
	return $r
}
proc ltmax { c m } {
	if {[llength $c] > [llength $m]} {
		set r 0
	} else {
		set r 1
	}
# puts stderr "ltmax ($c) ($m) -> $r"
	return $r
}
proc incrcnt { c m } {
	set r {}
	set done 0
	foreach ec $c em $m {
		if {! $done} {
			set e [expr ($ec + 1) % $em]
# puts stderr "incrcnt $ec ($c) $em ($m) -> $e ($r)"
			if {$e != 0} {
				set done 1
			}
		} else {
			set e $ec
		}
		lappend r $e
	}
# puts stderr "incrcnt ($c) ($m) -> ($r) done $done"
	if {! $done} {
		lappend r 1
	}
	return $r
}

proc batch_suite {depth rec} {
	global maininfo specstate
	set maininfo(chooseInputs) 1
	log_time_and_state
	set specstate(fid) [open /tmp/states.[pid] w]
	set d 0
	set startState [spec_get_super]
	set autStart [mkstate $startState 0]
	if {[string compare $startState ""] == 0} {
		fatal "cannot get super state\n"
	}
	set states($autStart) 1
	set todo($startState) 1
	
	set in 0
	set out 1
	set nin 0
	set nout 0
	
	set fout [open /tmp/suite.$out.[pid] w]
	set ttt [list PH $startState]
	puts $fout [join $ttt " ; "]
	incr nout
	close $fout
	
	set tmp $in
	set in $out
	set out $tmp
	unset tmp
	
	set nin $nout
	set nout 0
	
#	set work_in [list $tt]
	set fin [open /tmp/suite.$in.[pid] r]
	set fout [open /tmp/suite.$out.[pid] w]
	set work_out {}
	set ready {}
	
	while {($depth <= 0 || $d < $depth) && [main_go_on] && $nin > 0} {
		puts stderr "working on depth $d nin $nin"
		while {[gets $fin line] >= 0} {
			incr nin -1
			set arr [split $line \;]
			set tt {}
			foreach e $arr {
				lappend tt [string trim $e]
			}
			set pos [get_ph_pos $tt]
			puts stderr "  nin $nin nout $nout npos [llength $pos] d $d work_in el \"$tt\" (pos: $pos)"
			if {[llength $pos] == 0} {
				# lappend ready $tt
				puts [join $tt " ; "]
				# puts stderr "ready: \"$tt\""
			} else {
				set maxidx {}
				set revpos [lreverse $pos]
				foreach p $revpos {
					set S [lindex $tt [expr $p + 1]]
					if {![info exists repls($S)]} {
						set rep {}
						spec_set_super $S
						set inputs [spec_get_inputs]
						set outputs [spec_get_outputs]
						# puts stderr "      working on el \"$tt\" (S: $S) (in: $inputs) (out: $outputs)"
						if {$depth <= 0} {
							set r  [list P 0 ]
							# puts stderr "        repl_out += $r"
							lappend rep $r
						}
						set ttsub {}
						foreach o [lindex $outputs 1] {
							spec_set_super $S
							set chan_spec [lindex $o 0]
							set ev_spec [lindex $o 1]
							set susp_spec [lindex $o 2]
							set preds_spec [lindex $o 3]
							check_output $chan_spec $ev_spec $susp_spec $preds_spec
							set Snew [spec_get_super]
							if {$rec && [string compare $Snew $startState] == 0} {
								lappend ttsub [lindex $o 1] R 0
							} else {
								lappend ttsub [lindex $o 1] PH $Snew
							}
						}
						set r [concat O [llength [lindex $outputs 1]] $ttsub]
						# puts stderr "        repl_out += $r"
						lappend rep $r
						foreach i  [lindex $inputs 1] {
							spec_set_super $S
							set chan_spec [lindex $i 0]
							set ev_spec [lindex $i 1]
							set susp_spec [lindex $i 2]
							set preds_spec [lindex $i 3]
							if {[string first var_ $ev_spec] >= 0 || [string compare $preds_spec ""] != 0 } {
								puts stderr "\tinstantiating  $ev_spec $susp_spec $preds_spec"
								set inst_result [instantiate_all $ev_spec $susp_spec $preds_spec]
								switch -- [lindex $inst_result 0] \
								  inst_ok {
									foreach res  [lindex $inst_result 1] {
										set inst_ev_spec [lindex $res 0]
										set inst_susp_spec [lindex $res 1]
										set inst_preds_spec [lindex $res 2]
										puts stderr "\tcheck_input $chan_spec $inst_ev_spec $inst_susp_spec $inst_preds_spec"
										spec_set_super $S
										check_input $chan_spec $inst_ev_spec $inst_susp_spec $inst_preds_spec
										set Snew [spec_get_super]
										if {$rec && [string compare $Snew $startState] == 0} {
											set r [list S 1 $inst_ev_spec R 0]
										} else {
											set r [list S 1 $inst_ev_spec PH $Snew]
										}
										# puts stderr "        repl_out += $r"
										lappend rep $r
									}
								} inst_error {
								}
							} else {
								check_input $chan_spec $ev_spec $susp_spec $preds_spec
								set Snew [spec_get_super]
								if {$rec && [string compare $Snew $startState] == 0} {
									set r [list S 1 [lindex $i 1] R 0]
								} else {
									set r [list S 1 [lindex $i 1] PH $Snew]
								}
								# puts stderr "        repl_out += $r"
								lappend rep $r
							}
						}
						set repls($S) $rep
						unset r
						unset rep
					}
					lappend maxidx [llength $repls($S)]
				}
				set cnts [initcount $maxidx]
				while {[ltmax $cnts $maxidx]} {
					set ttt $tt
					foreach p $revpos c $cnts {
						set S [lindex $ttt [expr $p + 1]]
						set r [lindex $repls($S) $c]
						set cmd [list lreplace $ttt $p [expr $p + 1]]
						foreach e $r {
							lappend cmd $e
						}
						set ttt [eval $cmd]
						# puts stderr "    p $p S $S c $c r $r"
						unset cmd
					}
					set pp [get_ph_pos $ttt]
					if {[llength $pp] == 0} {
						# lappend ready $ttt
						puts [join $ttt " ; "]
						# puts stderr "ready: \"$ttt\""
					} else {
						# lappend work_out $ttt
						puts $fout [join $ttt " ; "]
						incr nout
					}
					# puts stderr "   ($cnts):($maxidx)     cur_out += $ttt"
					unset ttt
					# puts stderr "precnts ($cnts) ($maxidx)"
					set cnts [incrcnt $cnts $maxidx]
					# puts stderr "postcnts ($cnts) ($maxidx)"
				}
#				foreach e $cur_out {
#					set pp [get_ph_pos $e]
#					if {[llength $pp] == 0} {
#						# lappend ready $e
#						puts [join $e " ; "]
#						# puts stderr "ready: \"$e\""
#					} else {
#						lappend work_out $e
#					}
#				}
			}
		}
		close $fin
		close $fout
		
		set tmp $in
		set in $out
		set out $tmp
		unset tmp

		set nin $nout
		set nout 0
	

		set fin [open /tmp/suite.$in.[pid] r]
		set fout [open /tmp/suite.$out.[pid] w]
	
#		unset work_in
#		set work_in $work_out
#		unset work_out
#		set work_out {}
		set d [expr $d + 1]
	}
	puts stderr "after:"
		close $fout
 		puts stderr "working on depth $d nin $nin"
		set r [list P 0]
		while {[gets $fin line] >= 0} {
			incr nin -1
			set arr [split $line \;]
			set tt {}
			foreach e $arr {
				lappend tt [string trim $e]
			}
			set pos [get_ph_pos $tt]
			puts stderr "  nin $nin nout $nout npos [llength $pos] d $d work_in el \"$tt\" (pos: $pos)"
			if {[llength $pos] == 0} {
				# lappend ready $tt
				puts [join $tt " ; "]
				# puts stderr "ready: \"$tt\""
			} else {
				set maxidx {}
				set revpos [lreverse $pos]
				set cnts [initcount $maxidx]
				while {[ltmax $cnts $maxidx]} {
					set ttt $tt
					foreach p $revpos c $cnts {
						set S [lindex $ttt [expr $p + 1]]
						set cmd [list lreplace $ttt $p [expr $p + 1]]
						foreach e $r {
							lappend cmd $e
						}
						set ttt [eval $cmd]
						# puts stderr "    p $p S $S c $c r $r"
						unset cmd
					}
					puts [join $ttt " ; "]
					# puts stderr "   ($cnts):($maxidx)     cur_out += $ttt"
					unset ttt
					# puts stderr "precnts ($cnts) ($maxidx)"
					set cnts [incrcnt $cnts $maxidx]
					# puts stderr "postcnts ($cnts) ($maxidx)"
				}
			}
		}
		close $fin
	spec_set_super $startState
	catch {send_error "End of Batch Test\n"}
}


# =========================================================================
