#! @PERL@ --
#
# $Id: iochooser.pl,v 1.2.2.2 2001/11/15 15:42:12 belinfan Exp $
#
# usage: iochooser [ -s seed ] [ -debug ] [ bounds_and_values ]
#
# This module of the CdR testing tool environment is meant as a filter between
# Driver and Primer:
# it reads the messages that are sent from Driver to Primer, and provides
# suggestions for C_IOKIND questions from the Driver.
# The suggestions are chosen randomly;
# the seed of the random number generator can be set using the (-s seed) flag.
#
# The bounds_and_values argument has the following form:
#
#      0.5:iokind=input:1.0:iokind=output
#
# i.e. it is a string containing pairs of bounds and values, separated by
# colons(:), that will be used in choose_value routine below,
# in the following way:
# The random number will be a value between 0 and 1;
# the bounds_and_values string will be scanned from left to right, and
# the value of the first bound that is < the chosen number will be returned.
# If no bound is smaller than the chosen random number, the empty string
# will be returned.
# The result of the choose_value routine will be sent as suggestion to
# the Primer. This implies that the values should be ``valid'' according
# to the Primer-Driver interface.
#
# NOTE that this interface has changed: in the PAST we used
#
#     0.5:A_IO_INPUT:1.0:A_IO_OUTPUT
#
# where we NOW use 
#
#      0.5:iokind=input:1.0:iokind=output
#
# instead.
# In the future (for MIOCO) we probably also want to specify a channel, as in
#
#      0.5:iokind=input<TAB>channel=in:1.0:iokind=output<TAB>channel=out
#
# or even use different values for different input or output channels, as in
# (the two lines below should be read as being a single line)
#
#      0.4:iokind=input<TAB>channel=in1:0.5:iokind=input<TAB>channel=in2:\
#         1.0:iokind=output<TAB>channel=out
#
# where <TAB> is the literal tab character.
# NOTE Probably it would be nice to be able to use \t instead of
# the literal tab character.
#

# What do we want to do?
# We initialise:
#  - We initialise the random number generator
# We go to our main loop:
#  - We read lines from standard input.
#     - If the start of a line matches the C_IOKIND command without suggestion
#       - we generate a suggestion
#       - if we have done this, we output the line (with suggestion)
#     - If a line does not match a C_IOKIND, we echo it on stdout
#
# ======================================

use strict;

require 5.002;
use Getopt::Long;

# ======================================

my (
	$prog,
	%optcl, $result, $debug, $seed,
	$bounds_and_values, @bounds_and_values
   );

# ======================================

$prog = 'iochooser';
$| = 1;
$seed = 0;
$bounds_and_values = "0.5:iokind=input:1.0:iokind=output";

# ======================================

%optcl = ();
$result = GetOptions(
			\%optcl,
			"s=i",
			"debug:i",
		);

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});
$seed = $optcl{"s"} if (defined $optcl{"s"});
$bounds_and_values = $ARGV[0] if $ARGV[0];

# ======================================

init_random($seed);
init_bounds_and_values($bounds_and_values);
main_loop();

# ======================================

sub init_random {
	my($seed) = @_;
	srand($seed);
}

# ======================================

sub init_bounds_and_values {
	my($bounds_and_values) = @_;
	@bounds_and_values = split(':',$bounds_and_values);
}

# ======================================

sub main_loop {
	my($line,$resultline,$lineprefix,$suggestion);
	while(<STDIN>) {
		$resultline = $_;
		chomp;
		print STDERR "  --- line = ($_)\n"
			if $debug;
		if (/(^C_IOKIND\s*)$/i) {
			$lineprefix = $1;
			$suggestion = choose_value(@bounds_and_values);
			$resultline =~ s/$lineprefix/$lineprefix $suggestion/;
			$_ = $resultline;
		}
		print STDERR "  --- resultline = ($resultline)\n"
			if $debug;
		print $resultline;
	}
}

# ======================================

sub choose_value {
	my(@bounds_and_values) = @_;
	my($nr,$result);
	$nr = rand(1);
	print STDERR "   --- random nr = ($nr)\n"
		if $debug;
	while( @bounds_and_values >= 1 ) {
		my($bound) = shift(@bounds_and_values);
		my($value) = shift(@bounds_and_values);
		if ($nr < $bound) {
			print STDERR "    --- random nr $nr <  bound $bound;          value = ($value)\n"
				if $debug;
			return $value;
		} else {
			print STDERR "    --- random nr $nr >= bound $bound; skipping value   ($value)\n"
				if $debug;
		}
	}
	$result = '';
	return $result;
}

# ======================================
