#! @TCLSH@
#! @EXPECT@ --
#! /Utils/bin/tclsh
#! /usr/local/bin/expect --

# NOTE: this script only needs tcl(sh); it does not use expect
# features. The only reason that we use the expect interpreter
# here is that the configure script of TorX finds the location of
# expect already for us, and (AFAIK) does not (yet?) look for tclsh.

# however, somehow on windows xp cygwin tclsh seems to work better.

proc find_user {host} {
	global env

	if {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set fname [file join $dir "$pfx-$user-$display"]
	regsub -all : $fname _ fname
	return $fname
}

proc try_connect { fname do_warn } {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			close $fid
			set words [split $line]
			set ipNr [lindex $words 0]
			if {[string compare $ipNr 0.0.0.0] == 0} {
				set ipNr 127.0.0.1
			}
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			if {![catch {socket $ipNr $portNr} socket]} {
				return $socket
			} elseif {$do_warn} {
				warning "could not connect tcp!$ipNr!$portNr: $socket"
			}
		} else {
			close $fid
		}
	} elseif {$do_warn} {
		warning "$fid"
	}
	return 0
}

# exp_internal 1
set debug 0

set prog mctrl

# we use output to make sure that everything is flushed!
# we do this to avoid deadlocks due to buffers being full...
proc output { c  s } {
	if {[catch {
		puts $c [string trim $s]
		flush $c
	} msg]} {
		catch {puts stderr "output: $msg"}
	}
}

proc fatal {s {code {0}}} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
	cleanup_exit $code
}
proc message {s} {
	global prog
	set s [string trim $s]
	output stdout "$prog: $s"
}
proc warning {s} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
}

proc debug {s} {
	global prog
	global debug
	if {$debug} {
		set s [string trim $s]
		output stderr "$prog: $s"
	}
}

proc cleanup_exit {{code {0}}} {
	global state
	if {[info exists state(stderr)]} {
		eval fconfigure stderr $state(stderr)
	}
	if {[info exists state(stdout)]} {
		eval fconfigure stdout $state(stdout)
	}
	if {[info exists state(stdin)]} {
		eval fconfigure stdin $state(stdin)
	}
	exit $code
}

proc configure_fids {socket} {
	fconfigure $socket -buffering line  -translation {binary binary}
	fconfigure stdin -buffering line \
		                   -translation {binary binary}
}

proc filevent_fids {socket remoteHost remotePort waitRemoteClose} {
	global forever
	fileevent $socket readable \
	    	[list getRemote $remoteHost $remotePort $socket]
	fileevent stdin readable \
	 	[list sendRemote $remoteHost $remotePort $socket $waitRemoteClose]
	if {[catch {vwait forever} msg]} {
		fatal "internal error in filevent_fids: $msg" 1
	}
}


proc getRemote {remoteHost remotePort sock} {
   if {[catch {
	fileevent $sock readable {}
	set read [gets $sock line]
	debug "getRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
			catch {close stdout}
 			cleanup_exit
	} else {
		output stdout $line
		debug "getRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent $sock readable [list getRemote $remoteHost $remotePort $sock]
    } msg]} {
      catch {close stdout }
      cleanup_exit
    }
}
proc sendRemote {remoteHost remotePort sock waitRemoteClose} {
  if {[catch {
	fileevent stdin readable {}
	if {[catch {gets stdin line} read]} {
		warning "sendRemote tcp!$remoteHost!$remotePort error?: $read"
		cleanup_exit
	}
	debug "sendRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
			# set line "A_QUIT"
			# catch {output $sock $line}
			debug "eof stdin"
			if {! $waitRemoteClose} {
				catch {close $sock}
				debug "sendRemote closed tcp!$remoteHost!$remotePort"
				cleanup_exit
			} else {
				return
			}
	} else {
		output $sock $line
		debug "sendRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent stdin readable [list sendRemote $remoteHost $remotePort $sock $waitRemoteClose]
   } msg]} {
			debug "eof stdin"
			if {! $waitRemoteClose} {
				catch {close $sock}
				debug "sendRemote closed tcp!$remoteHost!$remotePort"
				cleanup_exit
			} else {
				return
			}
   }
}

proc main_loop {sock waitRemoteClose argstring} {
	configure_fids $sock
	set remote [fconfigure $sock -peername]
	set remoteHost [lindex $remote 0]
	set remotePort [lindex $remote 2]
	catch {
		puts $sock $argstring
		flush $sock
	}
	filevent_fids $sock $remoteHost $remotePort $waitRemoteClose
}

proc main {} {
	global argv
	global env

	set waitRemoteClose 0

	set reuse ""
	set mcast ""
	set title ""
	set key ""
	set files {}
	set exit 0

	set argstring ""
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast "-m $env(TORXMCASTID)"
	}

	set args $argv
	
#	set joinedargs [join $argv ")(" ]
#	warning "command lin args: ($joinedargs)"

	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse "-r"
		  } -m {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set mcast "-m [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set title "-t {[lindex $args 0]}"
					set args [lrange $args 1 end]
				}
		  } -k {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set key "-k [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -exit {
		  		set exit 1
		  } -* {
		  		warning "unknown flag: $a"
		  } default {
		  		lappend files $a
		  }
		}
	}

	if {$exit || ([llength $files] == 0)} {
		set host [exec uname -n]
		set display [find_display $host]
		set user [find_user $host]
		set fname [mk_filename /tmp mctrl $user $display]
		# puts stderr "fname=$fname"

		set sock [try_connect $fname 0]

		set argstring "-args $reuse $mcast $title"

		if {[string compare $sock 0] != 0} {
			if {$exit} {
				catch {puts $sock -exit}
				exit
			}
			# puts stderr "sending $argstring"
			main_loop $sock $waitRemoteClose $argstring
			return
		} elseif {! $exit} {
			start mctrlsrv
			set i 0
			set maxtries 10
			while {$i < $maxtries} {
				if {$i == ($maxtries - 1)} {
					set sock [try_connect $fname 1]
				} else {
					set sock [try_connect $fname 0]
				}
				if {[string compare $sock 0] != 0} {
					main_loop $sock $waitRemoteClose $argstring
					return
				}
				after 2000
				incr i
			}
		}
	} else {
		warning "usage: mctrl \[-r\] \[-t title\] \[-m mcastid\]"
		exit 1
	}
}

proc start { argv } {
	exec sh -c "expnohup $argv &" &
}

main
